/*
   Module: sineqm.c  - example TINE sine generator server, which illustrates
           the basic server API
*/
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "errors.h"
#include "rpctypes.h"
#include "servdef.h"
#include "swaplib.h"  /* for CF_STRUCT example */
#include "almsrv.h"
#include "toolkit.h" 

#ifndef MSDOS
# ifdef UNIX
# include "unixcmd.h"
# endif
# if defined(UNIX) || defined(WINNT)
# define random(r) ((r) != 0 ? rand()%(r) : 0)
# else
# define random(r) ((r) != 0 ? (i*(69069))%(r) : 0)
# endif
#else
# include "doscmd.h"
#endif

#define PRP_SINE      1
#define PRP_AMPLITUDE 2
#define PRP_FREQUENCY 3
#define PRP_PHASE     4
#define PRP_NOISE     5
#define PRP_INFO      6

#define NUM_DEVICES 10
#define NUM_VALUES (1024*8)

/*
   The code section below illustrates how one registers a structure
   to be exported.  
*/

typedef struct
{
  float amplitude;
  float frequency;
  float noise;
  float phase;
  int   numberCalls;
  char  description[64];
} SineInfo;

#define quit(i) { printf("Register struct: out of memory\n"); exit(i); }
void registerStructs(void)
{
  /* this must follow the order of the structure explicitly! */
  if (fillstruct("SineInfo",OFFSETIN(SineInfo,amplitude),1,CF_FLOAT)) quit(1);
  if (fillstruct("SineInfo",OFFSETIN(SineInfo,frequency),1,CF_FLOAT)) quit(1);
  if (fillstruct("SineInfo",OFFSETIN(SineInfo,noise),1,CF_FLOAT)) quit(1);
  if (fillstruct("SineInfo",OFFSETIN(SineInfo,phase),1,CF_FLOAT)) quit(1);
  if (fillstruct("SineInfo",OFFSETIN(SineInfo,numberCalls),1,CF_LONG)) quit(1);
  if (fillstruct("SineInfo",OFFSETIN(SineInfo,description),64,CF_TEXT)) quit(1);
  /* terminate the structure definition like this! */
  if (fillstruct("SineInfo",sizeof(SineInfo),NUM_DEVICES,CF_NULL)) quit(1);
}

SineInfo sineInfoTable[NUM_DEVICES];
float sinbuf[NUM_DEVICES][NUM_VALUES];


void updateSine(int devnr)
{ /* used internally -> don't bother with bounds checking */
  int i, k;
  float *target = sinbuf[devnr];
  float f = sineInfoTable[devnr].frequency;
  int n = (int)sineInfoTable[devnr].noise;
  int p = (int)sineInfoTable[devnr].phase;
  float a = sineInfoTable[devnr].amplitude;

  for (i=0; i<NUM_VALUES; i++)
  {
    k = (i + p) % NUM_VALUES;
    target[k] = random(n) + (float)(sin(f*i*6.2832/(NUM_VALUES/8))*a);
  }
  return;
}
char *sinname[NUM_DEVICES] = {
  "SineGen0","SineGen1","SineGen2","SineGen3","SineGen4",
  "SineGen5","SineGen6","SineGen7","SineGen8","SineGen9" };

void sininit(void)
{
  int i;

  /* some initial values for all sine devices */
  for (i=0; i<NUM_DEVICES; i++)
  {
    sineInfoTable[i].amplitude = 256.0;
    sineInfoTable[i].frequency = 1.0;
    sineInfoTable[i].phase = 0.0;
    sineInfoTable[i].noise = 30.0;
    sprintf(sineInfoTable[i].description,"Sine Generator %d at your disposal",i);
  }
  registerStructs();

  /* register device names, bind to device number: */
  for (i=0; i<NUM_DEVICES; i++) RegisterDeviceName("SINEQM",sinname[i],i);
}

void sinbkg(void)
{
  static int ptr = 0;
  static int cnt = 0;
  int i;
  /* do IO or something useful here */
  for (i=0; i<NUM_DEVICES; i++)
  {
    updateSine(i);
  }
}
int sineqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,i,cc;
  float mcarray[NUM_DEVICES], fval;
  SineInfo *sinf;

  devnr = GetDeviceNumber("SINEQM",devName);
  if (devnr < 0 || devnr >= NUM_DEVICES) return illegal_equipment_number;

  prpid = GetPropertyId("SINEQM",devProperty);

  switch (prpid)
  {
    case PRP_SINE:
        if (access&CA_WRITE) return illegal_read_write;
        if ((cc=putValuesFromFloat(dout,sinbuf[devnr],NUM_VALUES)) != 0) return cc;
        sineInfoTable[devnr].numberCalls++;
        return 0;
    case PRP_AMPLITUDE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 1 || fval > 1000) return out_of_range;
          sineInfoTable[devnr].amplitude = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].amplitude;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_FREQUENCY:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 1 || fval > 100) return out_of_range;
          sineInfoTable[devnr].frequency = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].frequency;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_PHASE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 0 || fval > 512) return out_of_range;
          sineInfoTable[devnr].phase = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].phase;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_NOISE:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=getValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 0 || fval > 100) return out_of_range;
          sineInfoTable[devnr].noise = fval;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<NUM_DEVICES; i++) 
            mcarray[i] = sineInfoTable[i].noise;
          if ((cc=putValuesFromFloatEx(dout,mcarray,NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_INFO:
        if (din->dArrayLength)
        { /* allow an atomic set of all relevant values this way : */
          if (din->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(din->dTag,"SineInfo",8)) return invalid_structure_tag;
          sinf = (SineInfo *)(din->data.vptr);
          if (sinf->amplitude < 1 || sinf->amplitude > 1000) return out_of_range;
          if (sinf->frequency < 1 || sinf->frequency > 100) return out_of_range;
          if (sinf->phase < 0 || sinf->phase > 512) return out_of_range;
          if (sinf->noise < 0 || sinf->noise > 100) return out_of_range;
          sineInfoTable[devnr] = *sinf;
        }
        if (dout->dArrayLength)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,"SineInfo",8)) return invalid_structure_tag;
          if (dout->dArrayLength > NUM_DEVICES) dout->dArrayLength = NUM_DEVICES;
          sinf = (SineInfo *)(dout->data.vptr);
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            sinf[i] = sineInfoTable[(i+devnr)%NUM_DEVICES];
          }
        }
        return 0;
    default: 
        return illegal_property;
  }
}
void PreSystemInit(void)
{
# ifdef FIND_ENS_AND_EXPORTNAME
  char exp[20],ctx[20],fec[20];
  int port;
  FindNameServerOnNetwork();
  GetMyServerAddress("SINEQM",exp,ctx,fec,&port);
  RegisterFecInformation(fec,"TST",ctx,"Test Sine Curve",
    "My Office","none","me",(UINT16)port);
# endif
  SetCycleDelay(0);
}
void PostSystemInit(void)
{
  RegisterEquipmentModule(NULL,"SINEQM",0,sineqm,sininit,sinbkg,200,NULL);
  /* don't let people poll faster than 20 Hz */
  MinPollingRate = MaxPollingRate = 50;

}

