// =============================================================
// ==
// ==   electron / proton State Equipment Module: streqm.c
// ==
// =============================================================
#include "stdlib.h"
#include "math.h"
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include "prolog.h"
#include "servdef.h"
#include "srvip.h"
#include "archive.h"
#include "rpctypes.h"
#include "database.h"
#include "brdcast.h"
#include "almsrv.h"
#include "errors.h"
#include "toolkit.h"
#include "fecerr.h"
#include "fecdef.h"

// state variables:
extern int ramping, running, beam, lumi;
extern int Pramping, Eramping, Pbeam, Ebeam;
extern int Prunning, Erunning;
// get these from the globals later:
int ramping, running, beam, lumi;
int Pramping, Eramping, Pbeam, Ebeam;
int Prunning, Erunning;

int logproblem(char *text,...);

typedef struct
{
  char username[16];
  char useraddr[16];
} CALLER;
CALLER lastCallerToChangeState;

#define ELECTRON 1
#define PROTON   2
#define BOTH     3

#define STATEFILE "state.csv"
#define STATEINFOFILE "stateinfo"
#define PROCINFOFILE "procinfo"
#define MACHINFOFILE "machinfo"

typedef struct 
{
  char devname[16];
  char statetag[32];
  int  machine;
  time_t ltd;       // last time of declaration
  UINT32 ttd;       // total time at declared state
  UINT32 ttd_bc;    // total time at declared state with beam conditions
  UINT32 ttd_tc;    // total time at declared state with test conditions
  UINT32 tnd;       // total number of state declarations
} DECLSTATE;

DECLSTATE *declStates = NULL;
int ndeclstates = 0;
#define NMACHINETYPES 2
UINT32 declaredState[NMACHINETYPES] = {0,0};  // undefined

typedef struct 
{
  char procname[16];
  char procfile1[32];
  char procfile2[32];
  char rule[16];
  int  nstates1;
  int  nstates2;
  NAME16 *statenames1;
  NAME16 *statenames2;
  time_t ltd;       // last time of declaration
  UINT32 ttd;       // total time at declared state
  UINT32 ttd_bc;    // total time at declared state with beam conditions
  UINT32 ttd_tc;    // total time at declared state with test conditions
  UINT32 tnd;       // total number of state declarations
} MACHSTATE;

MACHSTATE *machStates = NULL;
int nmachstates = 0;

#define EBEAMCRITERIAFILE "EBeamCriteria.csv"
#define PBEAMCRITERIAFILE "PBeamCriteria.csv"
#define ETESTCRITERIAFILE "ETestCriteria.csv"
#define PTESTCRITERIAFILE "PTestCriteria.csv"

#define STOCK_STRINGLENGTH 128
typedef struct
{
  int configured;
  int nparams;
  int valid;
  int linkstatus;
  char tag[32];
  NAME64 *dserver;
  NAME32 *dproperty;
  DTYPE *param;
  DTYPE *threshold;
} VALIDITYCHECK; 

#define BEAMTYPE 0
#define TESTTYPE 1
#define FINDMTYPE(i) (((i) & 0x000f) - 1)
#define FINDCTYPE(i) ((i) >> 8)
#define MAKETYPEID(c,m) (((c) << 8) | (m+1))

VALIDITYCHECK isValidConditions[2*NMACHINETYPES];
VALIDITYCHECK *isBeamConditions = isValidConditions;
VALIDITYCHECK *isTestConditions = &isValidConditions[NMACHINETYPES];

void checkValidity(VALIDITYCHECK *vc);

int readValidityFile(char *fn, VALIDITYCHECK *vc)
{
  FILE *fp;
  DTYPE *param, *threshold;
  NAME64 *dserver;
  NAME32 *dproperty;
  int cc=0,done=0,prp_col,srv_col,fmt_col,siz_col,thr_col;
  char str[256],hdr[256],scratch[STOCK_STRINGLENGTH];
  int n=0,len,fmtsiz;

  vc->configured = FALSE;
  if ((n=GetDataBaseSize(fn)) <= 0) return -n;
  if ((dserver=(NAME64 *)calloc(n,sizeof(NAME64))) == NULL)
    return out_of_server_memory;
  if ((dproperty=(NAME32 *)calloc(n,sizeof(NAME32))) == NULL)
    return out_of_server_memory;
  if ((param=(DTYPE *)calloc(n,sizeof(DTYPE))) == NULL)
    return out_of_server_memory;
  if ((threshold=(DTYPE *)calloc(n,sizeof(DTYPE))) == NULL)
    return out_of_server_memory;

  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((srv_col=findcol(hdr,"SERVER")) < 0) ccerr(no_such_line);
      if ((prp_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(hdr,"FORMAT")) < 0) ccerr(no_such_line);
      if ((siz_col=findcol(hdr,"SIZE")) < 0) ccerr(no_such_line);
      if ((thr_col=findcol(hdr,"THRESHOLD")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(dserver[n].name,colptr(srv_col,str),64);
    strncpy(dproperty[n].name,colptr(prp_col,str),32);
    strncpy(scratch,colptr(fmt_col,str),16);
    param[n].dFormat = ftoi(scratch);
    threshold[n].dFormat = param[n].dFormat;
    fmtsiz = fmtsizeof(param[n].dFormat);
    strncpy(scratch,colptr(siz_col,str),16);
    param[n].dArrayLength = atoi(scratch);
    len =threshold[n].dArrayLength = param[n].dArrayLength;
    if ((param[n].data.vptr=calloc(len,fmtsiz)) == NULL)
      ccerr(out_of_server_memory);
    strncpy(scratch,colptr(thr_col,str),STOCK_STRINGLENGTH);
    switch (threshold[n].dFormat)
    {   
      case CF_FLOAT:
        if ((threshold[n].data.vptr=calloc(1,fmtsiz)) == NULL)
          ccerr(out_of_server_memory);
        threshold[n].data.fptr[0] = (float)strtod(scratch,NULL);
        break;
      case CF_DOUBLE:
        if ((threshold[n].data.vptr=calloc(1,fmtsiz)) == NULL)
          ccerr(out_of_server_memory);
        threshold[n].data.dptr[0] = strtod(scratch,NULL);
        break;
      case CF_LONG:
        if ((threshold[n].data.vptr=calloc(1,fmtsiz)) == NULL)
          ccerr(out_of_server_memory);
        threshold[n].data.lptr[0] = atoi(scratch);
        break;
      case CF_SHORT:
        if ((threshold[n].data.vptr=calloc(1,fmtsiz)) == NULL)
          ccerr(out_of_server_memory);
        threshold[n].data.sptr[0] = (short)atoi(scratch);
        break;
      case CF_TEXT:
      case CF_NAME64:
      case CF_NAME48:
      case CF_NAME32:
      case CF_NAME16:
      case CF_NAME8:
        if ((threshold[n].data.vptr=calloc(STOCK_STRINGLENGTH,1)) == NULL)
          ccerr(out_of_server_memory);
        scratch[STOCK_STRINGLENGTH-1] = 0; strupr(scratch);
        strncpy(threshold[n].data.cptr,scratch,STOCK_STRINGLENGTH);
        break;
      default:
        ccerr(illegal_format);
    }
    n++;
  }
  vc->nparams = n;
  vc->dserver = dserver;
  vc->dproperty = dproperty;
  vc->param = param;
  vc->threshold = threshold;
err:
  fclose(fp);
  return cc;
}
void vcallback(int id,int cc)
{
  VALIDITYCHECK *vc;
  int i;
  if (cc)
  {
    return;
  }
  for (i=0; i<2*NMACHINETYPES; i++) 
  {
    vc = &isValidConditions[i];
    if (vc->configured) checkValidity(vc);
  }
  return;
}
DTYPE *assignValidityParam(VALIDITYCHECK *vc,int id)
{
  int i,j;
  for (i=0; i<2*NMACHINETYPES; i++)
  {
    if (vc == &isValidConditions[i]) break;
    for (j=0; j<isValidConditions[i].nparams; j++)
    {
      if (strncmp(vc->dserver[id].name,isValidConditions[i].dserver[j].name,64))
        continue;
      if (strncmp(vc->dproperty[id].name,isValidConditions[i].dproperty[j].name,32))
        continue;
      if (vc->param[id].dFormat != isValidConditions[i].param[j].dFormat) continue;
      if (vc->param[id].dArrayLength != isValidConditions[i].param[j].dArrayLength) continue;

      return &isValidConditions[i].param[j];
    }
  }
  return NULL;
}
void startValidityLinks(VALIDITYCHECK *vc,int check_id)
{
  int i,id;
  char str[128];
  DTYPE *param;
  
  vc->configured = TRUE;
  vc->linkstatus = 0; /* hope for the best */
  for (i=0; i<vc->nparams; i++)
  {
    if ((param=assignValidityParam(vc,i)) != NULL)
    {
      vc->param[i] = *param;
      sprintf(str,"%s : multiple state criterion link to %s %s\n",
              vc->tag,vc->dserver[i].name,vc->dproperty[i].name);
      feclog(str);
      printf(str);
    }
    else
    {
      id = AttachLinkEx(vc->dserver[i].name,vc->dproperty[i].name,&vc->param[i],NULL,CA_READ,1000,vcallback,CM_POLL,check_id);
      if (id < 0)
      {
        sprintf(str,"%s : Cannot establish state criterion link to %s %s : %s\n",
                vc->tag,vc->dserver,vc->dproperty,erlst[-id]);
        feclog(str);
        logproblem(str);
        vc->linkstatus = -id;
      }
    }
  }
}
char *trimString(char *str)
{
  static char trimBuffer[256];
  char *pstr;
  if (str == NULL)
  {
    trimBuffer[0] = 0;
    goto err;
  }
  for (pstr=str; strchr(" \t\n",*pstr) != NULL; pstr++);
  strncpy(trimBuffer,str,256);
  strtok(trimBuffer," \t\n");
err:
  return trimBuffer;
}
void checkValidity(VALIDITYCHECK *vc)
{
  int isvalid = TRUE;
  int i;
  char tmpstr[STOCK_STRINGLENGTH],scratch[64];

  for (i=0; isvalid && i<vc->nparams; i++)
  {
    switch (vc->param[i].dFormat)
    {
      case CF_FLOAT: 
        if (vc->param[i].data.fptr[0] <
            vc->threshold[i].data.fptr[0]) isvalid = FALSE;
        break;
      case CF_DOUBLE: 
        if (vc->param[i].data.dptr[0] <
            vc->threshold[i].data.dptr[0]) isvalid = FALSE;
        break;
      case CF_LONG: 
        if (vc->param[i].data.lptr[0] <
            vc->threshold[i].data.lptr[0]) isvalid = FALSE;
        break;
      case CF_SHORT: 
        if (vc->param[i].data.sptr[0] <
            vc->threshold[i].data.sptr[0]) isvalid = FALSE;
        break;
      case CF_TEXT:
      case CF_NAME64:
      case CF_NAME48:
      case CF_NAME32:
      case CF_NAME16:
      case CF_NAME8:
        strncpy(scratch,vc->param[i].data.cptr,vc->param[i].dArrayLength);
        scratch[vc->param[i].dArrayLength] = 0;
        strupr(scratch);
        strcpy(tmpstr,trimString(scratch));
        if (strstr(vc->threshold[i].data.cptr,tmpstr) == NULL) isvalid = FALSE;
        break;
      default:
        isvalid = FALSE;
    }
  }
  vc->valid = isvalid;
  if (NGdebug > 0) 
  {
    sprintf(tmpstr,"%s conditions are %s",vc->tag,isvalid ? "valid" : "not valid");
    dbgoutput(tmpstr,tnull);
  }
}

typedef struct
{
  char devname[16];
  char structfile[32];
  int machine;
  int  nstates;
  time_t ltd;       // last time of declaration
  UINT32 tnd;       // total number of procedure declarations
  NAME16 *statenames;
} XDECLPROC;

XDECLPROC *declProcs = NULL;
int ndeclprocs = 0;

UINT32 declaredProc[NMACHINETYPES] = {0,0}; // undefined

int readMachStructureFile(char *fn,NAME16 **nl,int *nn)
{
  FILE *fp;
  int cc=0,done=0,sta_col;
  char str[256],hdr[256];
  int n=0;
  NAME16 *nms = NULL;
 
  if (nl == NULL || nn == NULL) return un_allocated;
  if ((n=GetDataBaseSize(fn)) <= 0) return -n;
  if ((nms=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL)
    return out_of_server_memory;
 
  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((sta_col=findcol(hdr,"STATE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(nms[n].name,colptr(sta_col,str),16);
    n++;
  }
err:
  fclose(fp);
  *nl = nms;
  *nn = n;
  return cc;
}
int readProcStructureFile(char *fn,XDECLPROC *dp)
{
  FILE *fp;
  int cc=0,done=0,sta_col;
  char str[256],hdr[256];
  int n=0;
  NAME16 *nms = NULL;

  if (dp == NULL) return un_allocated;
  if ((n=GetDataBaseSize(fn)) <= 0) return -n;
  if ((nms=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL)
    return out_of_server_memory;

  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((sta_col=findcol(hdr,"STATE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(nms[n].name,colptr(sta_col,str),16);
    n++;
  }
err:
  fclose(fp);
  dp->statenames = nms;
  dp->nstates = n;
  return cc;
}
int readProcInfoFromFile(void)
{
  FILE *fp;
  int cc=0,done=0,pro_col,str_col,mac_col,tnd_col,ltd_col;
  char str[256],hdr[256],scratch[64];
  int i,n=0,na=0,renameit=0;
  char fn[64],fna[64];

  sprintf(fn,"%s.csv",PROCINFOFILE);
  if ((n=GetDataBaseSize(fn)) <= 0) return -n;
  sprintf(fna,"%s.add",PROCINFOFILE);
  if ((na=GetDataBaseSize(fna)) > 0) n += na;
  if (n > ndeclprocs && declProcs != NULL) free(declProcs);
  if ((declProcs == NULL) && 
      (declProcs=(XDECLPROC *)calloc(n,sizeof(XDECLPROC))) == NULL)
    return out_of_server_memory;

  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((pro_col=findcol(hdr,"PROCEDURE")) < 0) ccerr(no_such_line);
      if ((str_col=findcol(hdr,"STRUCTUREFILE")) < 0) ccerr(no_such_line);
      if ((mac_col=findcol(hdr,"MACHINE")) < 0) ccerr(no_such_line);
      if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
      if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(declProcs[n].devname,colptr(pro_col,str),16);
    strncpy(declProcs[n].structfile,colptr(str_col,str),32);
    if ((cc=readProcStructureFile(declProcs[n].structfile,&declProcs[n])) != 0) ccerr(cc);
    strncpy(scratch,colptr(mac_col,str),16);
    declProcs[n].machine = atoi(scratch);
    strncpy(scratch,colptr(ltd_col,str),16);
    declProcs[n].ltd = atoi(scratch);
    strncpy(scratch,colptr(tnd_col,str),16);
    declProcs[n].tnd = atoi(scratch);
    n++;
  }
  if (na > 0)  // there's an .add file (new states!)
  {
    fclose(fp); done = 0;
    if ((fp=fopen(fna,"r")) == NULL) ccerr(file_error);
    while ( fgets(str,255,fp) != NULL)
    {
      if (strchr("%;#\n",str[0])) continue;
      if (!done)
      {
        strcpy(hdr,str);
        if ((pro_col=findcol(hdr,"PROCEDURE")) < 0) ccerr(no_such_line);
        if ((str_col=findcol(hdr,"STRUCTUREFILE")) < 0) ccerr(no_such_line);
        if ((mac_col=findcol(hdr,"MACHINE")) < 0) ccerr(no_such_line);
        if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
        if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }
      strncpy(scratch,colptr(pro_col,str),16);
      for (i=0; i<n && strncmp(scratch,declProcs[i].devname,16); i++);
      strncpy(declProcs[i].devname,scratch,16);
      strncpy(declProcs[i].structfile,colptr(str_col,str),32);
      if ((cc=readProcStructureFile(declProcs[i].structfile,&declProcs[i])) != 0)
        ccerr(cc);
      strncpy(scratch,colptr(mac_col,str),16);
      declProcs[i].machine = atoi(scratch);
      strncpy(scratch,colptr(ltd_col,str),16);
      declProcs[i].ltd = atoi(scratch);
      declProcs[i].ltd = atoi(scratch);
      strncpy(scratch,colptr(tnd_col,str),16);
      declProcs[i].tnd = atoi(scratch);
      if (i == n) n++;
    }
    renameit = TRUE;
  }
err:
  if (fp != NULL) fclose(fp);
  ndeclprocs = n;  
  if (na > 0 && renameit)
  {
    sprintf(fn,"%s.added",PROCINFOFILE);
    rename(fna,fn);
  }
  return cc;
}

int writeProcInfoToFile(void)
{
  FILE *fp;
  int i;
  char fn[64];

  sprintf(fn,"%s.csv",PROCINFOFILE);
  if ((fp=fopen(fn,"w")) == NULL) return file_error;
  fprintf(fp,"Procedure,StructureFile,Machine,LastTime,TotalNumber\n");
  for (i=0; i<ndeclprocs; i++)
  {
    fprintf(fp,"%s,%s,%ld,%ld,%ld\n",
            declProcs[i].devname,declProcs[i].structfile,
            declProcs[i].machine,declProcs[i].ltd,
            declProcs[i].tnd);
  }
  fclose(fp);
  return 0;
}
int writeStatesToFile(void)
{
  FILE *fp;

  if ((fp=fopen(STATEFILE,"w")) == NULL) return file_error;
  fprintf(fp,"ElectronState,ElectronCode,ProtonState,ProtonCode\n");
  fprintf(fp,"%s,%d,%s,%d\n",
          declStates[declaredState[0]].devname,declaredState[0],
          declStates[declaredState[1]].devname,declaredState[1]);
  fclose(fp);
  return 0;
}
int readStatesFromFile(void)
{
  FILE *fp;
  int cc=0,done=0,ec_col,pc_col;
  int i,j;
  char str[256],hdr[256],scratch[16];

  if ((fp=fopen(STATEFILE,"r")) == NULL) return file_error;
  while (fgets(str,255,fp) != NULL)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((ec_col=findcol(hdr,"ELECTRONCODE")) < 0) ccerr(no_such_line);
      if ((pc_col=findcol(hdr,"PROTONCODE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(ec_col,str),16);
    declaredState[0] = atoi(scratch);
    strncpy(scratch,colptr(pc_col,str),16);
    declaredState[1] = atoi(scratch);
    break;
  }
  for (i=0; i<ndeclprocs; i++) for (j=0; j<declProcs[i].nstates; j++)
  {
    if (!strncmp(declStates[declaredState[0]].devname,declProcs[i].statenames[j].name,16)) 
    {
      declaredProc[0] = i;
    }
    if (!strncmp(declStates[declaredState[1]].devname,declProcs[i].statenames[j].name,16)) 
    {
      declaredProc[1] = i;
    }
  }
err:
  fclose(fp);
  return cc;
}
int writeStateInfoToFile(void)
{
  FILE *fp;
  int i;
  char fn[64];

  sprintf(fn,"%s.csv",STATEINFOFILE);
  if ((fp=fopen(fn,"w")) == NULL) return file_error;
  fprintf(fp,"DeviceName,StateTag,Machine,LastTime,TotalTime,TotalTime_bc,TotalTime_tc,TotalNumber\n");
  for (i=0; i<ndeclstates; i++)
  {
    fprintf(fp,"%s,%s,%ld,%ld,%ld,%ld,%ld,%ld\n",
            declStates[i].devname,declStates[i].statetag,
            declStates[i].machine,declStates[i].ltd,
            declStates[i].ttd,declStates[i].ttd_bc,
            declStates[i].ttd_tc,declStates[i].tnd);
  }
  fclose(fp);
  return 0;
}
int readStateInfoFromFile(void)
{
  FILE *fp;
  int cc=0,done=0,dev_col,sta_col,mac_col,ltd_col,ttd_col,tnd_col;
  int ttdbc_col=0,ttdtc_col=0;
  char str[256],hdr[256],scratch[64];
  int i,n=0,na=0,renameit=0;
  char fn[64],fna[64];

  sprintf(fn,"%s.csv",STATEINFOFILE);
  if ((n=GetDataBaseSize(fn)) <= 0) return -n;
  sprintf(fna,"%s.add",STATEINFOFILE);
  if ((na=GetDataBaseSize(fna)) > 0) n += na;
  if (n > ndeclstates && declStates != NULL) free(declStates);
  if ((declStates == NULL) && 
      (declStates=(DECLSTATE *)calloc(n,sizeof(DECLSTATE))) == NULL)
    return out_of_server_memory;

  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((dev_col=findcol(hdr,"DEVICENAME")) < 0) ccerr(no_such_line);
      if ((sta_col=findcol(hdr,"STATETAG")) < 0) ccerr(no_such_line);
      if ((mac_col=findcol(hdr,"MACHINE")) < 0) ccerr(no_such_line);
      if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
      if ((ttd_col=findcol(hdr,"TOTALTIME")) < 0) ccerr(no_such_line);
      ttdbc_col = findcol(hdr,"TOTALTIME_BC");
      ttdtc_col = findcol(hdr,"TOTALTIME_TC");
      if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(declStates[n].devname,colptr(dev_col,str),16);
    strncpy(declStates[n].statetag,colptr(sta_col,str),32);
    strncpy(scratch,colptr(mac_col,str),16);
    declStates[n].machine = atoi(scratch);
    strncpy(scratch,colptr(ltd_col,str),16);
    declStates[n].ltd = atoi(scratch);
    strncpy(scratch,colptr(ttd_col,str),16);
    declStates[n].ttd = atoi(scratch);
    if (ttdbc_col >= 0)
    {
      strncpy(scratch,colptr(ttdbc_col,str),16);
      declStates[n].ttd_bc = atoi(scratch);
    }
    if (ttdtc_col >= 0)
    {
      strncpy(scratch,colptr(ttdtc_col,str),16);
      declStates[n].ttd_tc = atoi(scratch);
    }
    strncpy(scratch,colptr(tnd_col,str),16);
    declStates[n].tnd = atoi(scratch);
    n++;
  }
  if (na > 0)  // there's an .add file (new states!)
  {
    fclose(fp); done = 0;
    if ((fp=fopen(fna,"r")) == NULL) ccerr(file_error);
    while ( fgets(str,255,fp) != NULL)
    {
      if (strchr("%;#\n",str[0])) continue;
      if (!done)
      {
        strcpy(hdr,str);
        if ((dev_col=findcol(hdr,"DEVICENAME")) < 0) ccerr(no_such_line);
        if ((sta_col=findcol(hdr,"STATETAG")) < 0) ccerr(no_such_line);
        if ((mac_col=findcol(hdr,"MACHINE")) < 0) ccerr(no_such_line);
        if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
        if ((ttd_col=findcol(hdr,"TOTALTIME")) < 0) ccerr(no_such_line);
        ttdbc_col = findcol(hdr,"TOTALTIME_BC");
        ttdtc_col = findcol(hdr,"TOTALTIME_TC");
        if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }
      strncpy(scratch,colptr(dev_col,str),16);
      for (i=0; i<n && strncmp(scratch,declStates[i].devname,16); i++);
      strncpy(declStates[i].devname,scratch,16);
      strncpy(declStates[i].statetag,colptr(sta_col,str),32);
      strncpy(scratch,colptr(mac_col,str),16);
      declStates[i].machine = atoi(scratch);
      strncpy(scratch,colptr(ltd_col,str),16);
      declStates[i].ltd = atoi(scratch);
      strncpy(scratch,colptr(ttd_col,str),16);
      declStates[i].ttd = atoi(scratch);
      if (ttdbc_col >= 0)
      {
        strncpy(scratch,colptr(ttdbc_col,str),16);
        declStates[i].ttd_bc = atoi(scratch);
      }
      if (ttdtc_col >= 0)
      {
        strncpy(scratch,colptr(ttdtc_col,str),16);
        declStates[i].ttd_tc = atoi(scratch);
      }
      strncpy(scratch,colptr(tnd_col,str),16);
      declStates[i].tnd = atoi(scratch);
      if (i == n) n++;
    }
    renameit = TRUE;
  }
err:
  if (fp != NULL) fclose(fp);
  ndeclstates = n;  
  if (na > 0 && renameit)
  {
    sprintf(fn,"%s.added",STATEINFOFILE);
    rename(fna,fn);
  }
  return cc;
}
int writeMachineInfoToFile(void)
{
  FILE *fp;
  int i;
  char fn[64];

  sprintf(fn,"%s.csv",MACHINFOFILE);
  if ((fp=fopen(fn,"w")) == NULL) return file_error;
  fprintf(fp,"MachineProcedure,ProcedureFile1,ProcedureFile2,Rule,LastTime,TotalTime,TotalTime_bc,TotalTime_tc,TotalNumber\n");
  for (i=0; i<nmachstates; i++)
  {
    fprintf(fp,"%s,%s,%s,%s,%ld,%ld,%ld,%ld,%ld\n",
            machStates[i].procname,machStates[i].procfile1,
            machStates[i].procfile2,machStates[i].rule,
            declStates[i].ltd,
            machStates[i].ttd,machStates[i].ttd_bc,
            machStates[i].ttd_tc,machStates[i].tnd);
  }
  fclose(fp);
  return 0;
}
int readMachineInfoFromFile(void)
{
  FILE *fp;
  int cc=0,done=0,proc_col,file1_col,file2_col,rule_col,ltd_col,ttd_col,tnd_col;
  int ttdbc_col=0,ttdtc_col=0;
  char str[256],hdr[256],scratch[64];
  int i,n=0,na=0,renameit=0;
  char fn[64],fna[64];

  sprintf(fn,"%s.csv",MACHINFOFILE);
  if ((n=GetDataBaseSize(fn)) <= 0) return -n;
  sprintf(fna,"%s.add",MACHINFOFILE);
  if ((na=GetDataBaseSize(fna)) > 0) n += na;
  if (n > nmachstates && machStates != NULL) free(machStates);
  if ((machStates == NULL) && 
      (machStates=(MACHSTATE *)calloc(n,sizeof(MACHSTATE))) == NULL)
    return out_of_server_memory;

  if ((fp=fopen(fn,"r")) == NULL) return file_error;
  for (n=0; fgets(str,255,fp) != NULL;)
  {
    if (strchr("%;#\n",str[0])) continue;
    if (!done)
    {
      strcpy(hdr,str);
      if ((proc_col=findcol(hdr,"MACHINEPROCEDURE")) < 0) ccerr(no_such_line);
      if ((file1_col=findcol(hdr,"PROCEDUREFILE1")) < 0) ccerr(no_such_line);
      if ((file2_col=findcol(hdr,"PROCEDUREFILE2")) < 0) ccerr(no_such_line);
      if ((rule_col=findcol(hdr,"RULE")) < 0) ccerr(no_such_line);
      if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
      if ((ttd_col=findcol(hdr,"TOTALTIME")) < 0) ccerr(no_such_line);
      ttdbc_col = findcol(hdr,"TOTALTIME_BC");
      ttdtc_col = findcol(hdr,"TOTALTIME_TC");
      if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(machStates[n].procname,colptr(proc_col,str),16);
    strncpy(machStates[n].procfile1,colptr(file1_col,str),32);
    if ((cc=readMachStructureFile(machStates[n].procfile1,&machStates[n].statenames1,&machStates[n].nstates1)) != 0) ccerr(cc);
    strncpy(machStates[n].procfile2,colptr(file2_col,str),32);
    if ((cc=readMachStructureFile(machStates[n].procfile2,&machStates[n].statenames2,&machStates[n].nstates2)) != 0) ccerr(cc);
    strncpy(machStates[n].rule,colptr(rule_col,str),16);
    strncpy(scratch,colptr(ltd_col,str),16);
    machStates[n].ltd = atoi(scratch);
    strncpy(scratch,colptr(ttd_col,str),16);
    machStates[n].ttd = atoi(scratch);
    if (ttdbc_col >= 0)
    {
      strncpy(scratch,colptr(ttdbc_col,str),16);
      machStates[n].ttd_bc = atoi(scratch);
    }
    if (ttdtc_col >= 0)
    {
      strncpy(scratch,colptr(ttdtc_col,str),16);
      machStates[n].ttd_tc = atoi(scratch);
    }
    strncpy(scratch,colptr(tnd_col,str),16);
    machStates[n].tnd = atoi(scratch);
    n++;
  }
  if (na > 0)  // there's an .add file (new states!)
  {
    fclose(fp); done = 0;
    if ((fp=fopen(fna,"r")) == NULL) ccerr(file_error);
    while ( fgets(str,255,fp) != NULL)
    {
      if (strchr("%;#\n",str[0])) continue;
      if (!done)
      {
        strcpy(hdr,str);
        if ((proc_col=findcol(hdr,"MACHINEPROCEDURE")) < 0) ccerr(no_such_line);
        if ((file1_col=findcol(hdr,"PROCEDUREFILE1")) < 0) ccerr(no_such_line);
        if ((file2_col=findcol(hdr,"PROCEDUREFILE2")) < 0) ccerr(no_such_line);
        if ((rule_col=findcol(hdr,"RULE")) < 0) ccerr(no_such_line);
        if ((ltd_col=findcol(hdr,"LASTTIME")) < 0) ccerr(no_such_line);
        if ((ttd_col=findcol(hdr,"TOTALTIME")) < 0) ccerr(no_such_line);
        ttdbc_col = findcol(hdr,"TOTALTIME_BC");
        ttdtc_col = findcol(hdr,"TOTALTIME_TC");
        if ((tnd_col=findcol(hdr,"TOTALNUMBER")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }

      strncpy(scratch,colptr(proc_col,str),16);
      for (i=0; i<n && strncmp(scratch,machStates[i].procname,16); i++);
      strncpy(machStates[i].procname,colptr(proc_col,str),16);
      strncpy(machStates[i].procfile1,colptr(file1_col,str),32);
      strncpy(machStates[i].procfile2,colptr(file2_col,str),32);
      strncpy(machStates[i].rule,colptr(rule_col,str),16);
      strncpy(scratch,colptr(ltd_col,str),16);
      machStates[i].ltd = atoi(scratch);
      strncpy(scratch,colptr(ttd_col,str),16);
      machStates[i].ttd = atoi(scratch);
      if (ttdbc_col >= 0)
      {
        strncpy(scratch,colptr(ttdbc_col,str),16);
        machStates[i].ttd_bc = atoi(scratch);
      }
      if (ttdtc_col >= 0)
      {
        strncpy(scratch,colptr(ttdtc_col,str),16);
        machStates[i].ttd_tc = atoi(scratch);
      }
      strncpy(scratch,colptr(tnd_col,str),16);
      machStates[i].tnd = atoi(scratch);
      if (i == n) n++;
    }
    renameit = TRUE;
  }
err:
  if (fp != NULL) fclose(fp);
  nmachstates = n;  
  if (na > 0 && renameit)
  {
    sprintf(fn,"%s.added",MACHINFOFILE);
    rename(fna,fn);
  }
  return cc;
}

void staexi(void)
{
  writeStateInfoToFile();
  writeProcInfoToFile();
  writeMachineInfoToFile();
}
void stabkg(void)
{
  int i,j,k,deadband;
  int match1, match2;
  static time_t lasttime = 0;
  time_t t = time(NULL);
  VALIDITYCHECK *vc;
  
  ClearAlarm("STAEQM",0);

  deadband = lasttime > 0 ? (t - lasttime) : 1;
  lasttime = t;
  
  i = declaredState[0];
  declStates[i].ttd += deadband;
  if (isBeamConditions[0].valid) declStates[i].ttd_bc += deadband;
  if (isTestConditions[0].valid) declStates[i].ttd_tc += deadband;

  if ((int)declaredState[1] == i) goto out; // don't double count

  i = declaredState[1];
  declStates[i].ttd += deadband;
  if (isBeamConditions[1].valid) declStates[i].ttd_bc += deadband;
  if (isTestConditions[1].valid) declStates[i].ttd_tc += deadband;

out:  
  for (i=0; i<nmachstates; i++)
  {
    match1 = FALSE; match2 = FALSE;
    // check if file1 matches
    for (j=0; j<machStates[i].nstates1; j++)
    {
      k = declaredState[0];
      if (!strncmp(declStates[k].devname,machStates[i].statenames1[j].name,16)) break;
    }
    if (j < machStates[i].nstates1) match1 = TRUE;

    // check if file2 matches
    for (j=0; j<machStates[i].nstates2; j++)
    {
      k = declaredState[1];
      if (!strncmp(declStates[k].devname,machStates[i].statenames2[j].name,16)) break;
    }
    if (j < machStates[i].nstates2) match2 = TRUE;

    // check Rule (OR or AND)
    if (!match1 && !match2) continue;
    if (!strncmp(machStates[i].rule,"OR",2))
    {
      // count
      machStates[i].ttd += deadband;
      if (isBeamConditions[1].valid) machStates[i].ttd_bc += deadband;
      if (isTestConditions[1].valid) machStates[i].ttd_tc += deadband;
      continue;
    }
    if (!match1 || !match2) continue;
    if (!strncmp(machStates[i].rule,"AND",3))
    {
      // count
      machStates[i].ttd += deadband;
      if (isBeamConditions[1].valid) machStates[i].ttd_bc += deadband;
      if (isTestConditions[1].valid) machStates[i].ttd_tc += deadband;
      continue;
    }
  }
  if ((t%60) == 0) 
  {
    writeStateInfoToFile();
    writeProcInfoToFile();
    writeMachineInfoToFile();
  }

  for (i=0; i<2*NMACHINETYPES; i++) 
  {
    vc = &isValidConditions[i];
    if (vc->configured && vc->linkstatus != 0)
      SetAlarm("STAEQM",0,512,NULL);
  }
  return;  
}
time_t declStateTime = 0;
void staini(void)
{
  int i,cc;
  char str[128];
  
  remove("problems.log");

  if ((cc=readStateInfoFromFile()) != 0)
  {
    sprintf(str,"Cannot read state info file : %s\n",erlst[cc]);
    feclog(str);
    printf(str);
  }

  if ((cc=readProcInfoFromFile()) != 0)
  {
    sprintf(str,"Cannot read procedure info file : %s\n",erlst[cc]);
    feclog(str);
    printf(str);
  }
  
  if ((cc=readMachineInfoFromFile()) != 0)
  {
    sprintf(str,"Cannot read machine info file : %s\n",erlst[cc]);
    feclog(str);
    printf(str);
  }

  for (i=0; i<ndeclstates; i++)
    RegisterModuleName("STAEQM",declStates[i].devname,i);
  for (i=0; i<ndeclprocs; i++)
    RegisterModuleName("STAEQM",declProcs[i].devname,i+ndeclstates);    
  for (i=0; i<nmachstates; i++)
    RegisterModuleName("STAEQM",machStates[i].procname,i+ndeclstates+ndeclprocs);    
  AssignNumModules("STAEQM",ndeclstates+ndeclprocs+nmachstates);
  
  if ((cc=readStatesFromFile()) != 0)
  {
    sprintf(str,"Cannot read state file : %s\n",erlst[cc]);
    feclog(str);
    printf(str);
  }
    
  declStateTime = time(NULL);

  if ((cc=readValidityFile(EBEAMCRITERIAFILE, &isBeamConditions[0])) != 0)
  {
    sprintf(str,"Cannot read state validity file %s : %s\n",
            EBEAMCRITERIAFILE,erlst[cc]);
    feclog(str); printf(str);
  }
  else
  {
    strcpy(isBeamConditions[0].tag,"E Beam");
    startValidityLinks(&isBeamConditions[0],MAKETYPEID(0,0));
  }
  if ((cc=readValidityFile(PBEAMCRITERIAFILE, &isBeamConditions[1])) != 0)
  {
    sprintf(str,"Cannot read state validity file %s : %s\n",
            PBEAMCRITERIAFILE,erlst[cc]);
    feclog(str);
    printf(str);
  }
  else
  {
    strcpy(isBeamConditions[1].tag,"P Beam");
    startValidityLinks(&isBeamConditions[1],MAKETYPEID(0,1));
  }
  if ((cc=readValidityFile(ETESTCRITERIAFILE, &isTestConditions[0])) != 0)
  {
    sprintf(str,"Cannot read state validity file %s : %s\n",
            ETESTCRITERIAFILE,erlst[cc]);
    feclog(str);
    printf(str);
  }
  else
  {
    strcpy(isTestConditions[0].tag,"E Test");
    startValidityLinks(&isTestConditions[0],MAKETYPEID(1,0));
  }
  if ((cc=readValidityFile(PTESTCRITERIAFILE, &isTestConditions[1])) != 0)
  {
    sprintf(str,"Cannot read state validity file %s : %s\n",
            PTESTCRITERIAFILE,erlst[cc]);
    feclog(str);
    printf(str);
  }
  else
  {
    strcpy(isTestConditions[1].tag,"P Test");
    startValidityLinks(&isTestConditions[1],MAKETYPEID(1,1));
  }

  feclog("state server initialized");
}
void trgCallback(int id,int cc)
{
  if (cc) printf("TRIGGER : %s\n",erlst[cc]);
}
int sndTriggerToArchive(char *trg)
{
  DTYPE din;
  int cc;
  short sval = 1;
  char devName[32],str[256];

  din.dFormat = CF_SHORT;
  din.dArrayLength = 1;
  din.data.sptr = (UINT16 *)&sval;
  sprintf(devName,"/%s/STORAGE/%s",gDeviceContext,trg);
  if ((cc=AttachLink(devName,"TRIGGER",NULL,&din,CA_WRITE,1000,trgCallback,CM_SINGLE)) < 0)
  {
    sprintf(str,"TRIGGER %s error : %s",devName,erlst[-cc]);
    logproblem(str);
    feclog(str);
  }
  else
  {
    feclog("send trigger %s to archive",trg);
  }
  return cc < 0 ? -cc : 0;
}
int forwardTrigger(char *devsrv,char *trg)
{
  DTYPE din;
  int cc;
  short sval = 1;
  char devName[48],str[256];

  sprintf(devName,"%s/%s",devsrv,trg);
  din.dFormat = CF_SHORT;
  din.dArrayLength = 1;
  din.data.sptr = (UINT16 *)&sval;
  if ((cc=AttachLink(devName,"STATECHANGE",NULL,&din,CA_WRITE,1000,trgCallback,CM_SINGLE)) < 0)
    printf("TRIGGER %s error : %s\n",devName,erlst[-cc]);
  sprintf(str,"forward trigger to %s",devsrv);
  feclog(str);
  return cc < 0 ? -cc : 0;
}
int staeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc, i, j, k, n, type,len,newproc = FALSE;
  int dc = FALSE, bc = FALSE, tc = FALSE;
  short stateArray[7];
  short dsize,devNr,fmt;
  NAME16 un;
  char addr[32];
  BYTE ipxaddr[12];
  UINT32 ipaddr;
  char str[256];

  if (devName[0] == '#') devNr = atoi(&devName[1]);
  else devNr = GetModuleNumber("STAEQM",devName);
  
  fmt = dout->dFormat;
  dsize = (short)dout->dArrayLength;
  
  if (devNr < 0) return illegal_equipment_number;

  if (!strcmp(devProperty,"ARCHSTATE"))  // deduced machine state(s)
  {
    if (access & CA_WRITE) return illegal_read_write;
    stateArray[6] = Eramping;
    stateArray[5] = Ebeam;
    stateArray[4] = Erunning;
    stateArray[3] = Pramping;
    stateArray[2] = Pbeam;
    stateArray[1] = Prunning;
    stateArray[0] = lumi;
    return putValuesFromShort(dout,stateArray,7);
  }
  if (!strcmp(devProperty,"CALLER"))  // deduced machine state(s)
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dout->dArrayLength < 1 || dout->dArrayLength > 2) return dimension_error;
    switch (dout->dFormat)
    {
      case CF_NAME16: len = 16; n = dout->dArrayLength; break;
      case CF_NAME32: len = 32; n = dout->dArrayLength; break;
      case CF_TEXT:   if (dout->dArrayLength < 16) return dimension_error;
                      len = 16; n = dout->dArrayLength/16; break;
      default:        return illegal_format;
    }
    strncpy(dout->data.cptr,lastCallerToChangeState.username,16);
    if (n > 1) strncpy(dout->data.cptr+len,lastCallerToChangeState.useraddr,16);
    return 0;
  }
  if (!strcmp(devProperty,"TIMESTATS"))  // state time/duration information
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devNr >= ndeclstates) return illegal_equipment_number;
    if (dout->dFormat != CF_LONG) return illegal_format;
    switch (dout->dArrayLength)
    {
      case 5:  dout->data.lptr[4] = declStates[devNr].ttd_tc;
      case 4:  dout->data.lptr[3] = declStates[devNr].ttd_bc;
      case 3:  dout->data.lptr[2] = declStates[devNr].tnd;
      case 2:  dout->data.lptr[1] = declStates[devNr].ltd;
      case 1:  dout->data.lptr[0] = declStates[devNr].ttd;
               break;
      default: return dimension_error;
    }
    return 0;
  }
  if ((dc=!strcmp(devProperty,"TIMECOUNTER")) ||  // state duration counter
      (bc=!strcmp(devProperty,"BEAMCOUNTER")) ||
      (tc=!strcmp(devProperty,"TESTCOUNTER")) )
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devNr >= ndeclstates + ndeclprocs + nmachstates) return illegal_equipment_number;
    if (dout->dFormat != CF_LONG) return illegal_format;
    if (devNr < ndeclstates)
    {
      for (i=devNr; i<(int)dout->dArrayLength+devNr; i++)
      {
        if (dc) dout->data.lptr[i-devNr] = declStates[i%ndeclstates].ttd;
        else if (bc) dout->data.lptr[i-devNr] = declStates[i%ndeclstates].ttd_bc;
        else if (tc) dout->data.lptr[i-devNr] = declStates[i%ndeclstates].ttd_tc;
      }
    }
    else if (devNr < ndeclstates + ndeclprocs)
    {
      for (i=devNr; i<(int)dout->dArrayLength+devNr && i < ndeclstates+ndeclprocs; i++)
      {
        dout->data.lptr[i-devNr] = 0;
        for (j=0; j<declProcs[i-ndeclstates].nstates; j++)
        {
          k = GetModuleNumber("STAEQM",declProcs[i-ndeclstates].statenames[j].name);
          if (k >= 0)
          {
            if (dc) dout->data.lptr[i-devNr] += declStates[k].ttd;
            else if (bc) dout->data.lptr[i-devNr] += declStates[k].ttd_bc;
            else if (tc) dout->data.lptr[i-devNr] += declStates[k].ttd_tc;
          }
        }
      }
    }
    else if (devNr < ndeclstates + ndeclprocs + nmachstates)
    {
      n = ndeclstates + ndeclprocs;
      for (i=devNr; i<(int)dout->dArrayLength+devNr; i++)
      {
        if (dc) dout->data.lptr[i-devNr] = machStates[i%n].ttd;
        else if (bc) dout->data.lptr[i-devNr] = machStates[i%n].ttd_bc;
        else if (tc) dout->data.lptr[i-devNr] = machStates[i%n].ttd_tc;
      }
    }
    return 0;
  }
  if (!strcmp(devProperty,"EVENTCOUNTER"))  // state number counter
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (devNr >= ndeclstates + ndeclprocs) return illegal_equipment_number;
    if (dout->dFormat != CF_LONG) return illegal_format;
    if (devNr < ndeclstates)
    {
      for (i=devNr; i<(int)dout->dArrayLength+devNr; i++)
      {
        dout->data.lptr[i-devNr] = declStates[i%ndeclstates].tnd;
      }
      setDataTimeStamp(declStates[devNr%ndeclstates].ltd);
    }
    else if (devNr < ndeclstates + ndeclprocs)
    {
      for (i=devNr; i<(int)dout->dArrayLength+devNr; i++)
      {
        dout->data.lptr[i-devNr] = declProcs[(i-ndeclstates)%ndeclprocs].tnd;
      }
      setDataTimeStamp(declProcs[(devNr-ndeclstates)%ndeclprocs].ltd);
    }
    return 0;
  }
  if (!strcmp(devProperty,"DECLSTATE"))  // declared machine state
  {
    if ((access&CA_WRITE)) return illegal_read_write;
    if (dout->dArrayLength != 2) return dimension_error;
    setDataTimeStamp(declStateTime);
    switch (dout->dFormat)
    {
      case CF_NAME16: 
           strncpy(dout->data.cptr,declStates[declaredState[0]].statetag,16);
           strncpy(&dout->data.cptr[16],declStates[declaredState[1]].statetag,16);
           return 0;          
      case CF_NAME32:
           strncpy(dout->data.cptr,declStates[declaredState[0]].statetag,32);
           strncpy(&dout->data.cptr[32],declStates[declaredState[1]].statetag,32);
           return 0;          
      default: return putValuesFromLong(dout,declaredState,2);
    }
  }
  if (!strcmp(devProperty,"STATELIST"))  // list machine states
  {
    if ((access&CA_WRITE)) return illegal_read_write;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        break;
      default: 
        return illegal_format;
    }
    len = fmtsizeof(dout->dFormat);
    n = MIN(32,len);
    for (i=0; i<(int)dout->dArrayLength && i<ndeclstates; i++)
    {
      strncpy(dout->data.cptr+len*i,declStates[i].statetag,n);
    }
    dout->dArrayLength = i;
    return 0;
  }
  if (!strcmp(devProperty,"DECLPROC"))  // declared machine state
  {
    if ((access&CA_WRITE)) return illegal_read_write;
    if (dout->dArrayLength != 2) return dimension_error;
    setDataTimeStamp(declStateTime);
    switch (dout->dFormat)
    {
      case CF_NAME16: 
           strncpy(dout->data.cptr,declProcs[declaredProc[0]].devname,16);
           strncpy(&dout->data.cptr[16],declProcs[declaredProc[1]].devname,16);
           return 0;          
      case CF_NAME32:
           strncpy(dout->data.cptr,declProcs[declaredProc[0]].devname,16);
           strncpy(&dout->data.cptr[32],declProcs[declaredProc[1]].devname,16);
           return 0;          
      default: return putValuesFromLong(dout,declaredProc,2);
    }
  }
  if (!strcmp(devProperty,"STATECHANGE"))  // deduced machine state(s)
  {
    if (!(access&CA_WRITE)) return illegal_read_write;
    if (devNr >= ndeclstates) return illegal_equipment_number;
    declStates[devNr].ltd = time(NULL);
    declStates[devNr].tnd++;
    type = declStates[devNr].machine;
    if (type&ELECTRON) declaredState[0] = devNr;
    if (type&PROTON) declaredState[1] = devNr;
    //printf("CHANGESTATE: %s %s (%d)\n>",devName,declStates[devNr].statetag,devNr);
    // cc = getValuesAsLong(din,&declaredState,1);
    // use it to trigger archives here ...
    declStateTime = time(NULL);
    writeStatesToFile();
    writeStateInfoToFile(); 
    sndTriggerToArchive(declStates[devNr].devname);
    // put caller into caller buffer:
    n = 1; GetCallerInfo(&un,ipxaddr,&ipaddr,NULL,&n);
    if (ipaddr != 0) inet_ntoa_b(*(struct in_addr *)&ipaddr,addr);
#   if defined(IPXBSD) || defined(IPXNW)
    else strncpy(addr,ipxnet_ntoa(ipxaddr),32);
#   endif
    strncpy(lastCallerToChangeState.username,un.name,16);
    strncpy(lastCallerToChangeState.useraddr,addr,16);
#   ifdef FORWARDING
    forwardTrigger("DataSrv",declStates[devNr].devname);
    forwardTrigger("DataSrv1",declStates[devNr].devname);
    forwardTrigger("HESTATEDATA",declStates[devNr].devname);
#   endif
    cc = dout->dArrayLength > 0 ? putValuesFromLong(dout,declaredState,2) : 0;
    sprintf(str,"State changed to %s by %s [%s]",
            declStates[devNr].statetag,un.name,addr);
    feclog(str);
    for (i=0; i<ndeclprocs; i++) for (j=0; j<declProcs[i].nstates; j++)
    {
      if (strncmp(devName,declProcs[i].statenames[j].name,16)) continue;
      if (type&ELECTRON)
      {
        if ((int)declaredProc[0] != i) { declaredProc[0] = i; newproc = TRUE; }
      }
      if (type&PROTON)
      {
        if ((int)declaredProc[1] != i) { declaredProc[1] = i; newproc = TRUE; }
      }
      if (newproc)
      {
        declProcs[i].ltd = declStateTime;
        declProcs[i].tnd++;
        writeProcInfoToFile();
      }
      return cc;
    }
    return cc;
  }
  return illegal_property;
}
void PreSystemInit(void)
{
  FeclogDepth = 2000;
  ListenForGlobals = 0;
}
void PostSystemInit(void)
{
  int cc;
  signal(SIGFPE,SIG_IGN); // catch floating point domain errors
  RequireAcknowledgments = 0; // turn off ack. requirments for REFRESH mode
  if ((cc=RegisterEquipmentModule(NULL,"STAEQM",0,staeqm,staini,stabkg,1000,staexi)) != 0)
    printf("Can't register STAEQM: %s\n",erlst[cc]);
} 
int logproblem(char *text,...)
{
  int len,cc=0;
  double ts = makeDataTimeStamp();
  char str[256];
  FILE *fp = NULL;
  va_list args;
  if (strlen(text) > 80) ccerr(string_too_long);
  va_start(args,text);
  sprintf(str,"%s : ",getDataTimeString(ts,0));
  vsprintf(&str[strlen(str)],text,args);
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  va_end(args);
  if ((fp=fopen("problems.log","a")) == NULL) ccerr(file_error);
  fwrite(str,len=strlen(str),1,fp);
err:
  if (fp) fclose(fp);
  return cc;
}
